#!/usr/local/bin/perl5.6.1 -w

# common code for Rancid "clogin" programs written in Perl
# interaction with a router - login, enable


package Rancid::Login;

use Expect;
use IO::Socket::INET;
use IO::Stty;

use Rancid::ParseConfig;

use strict;
use warnings;
use Carp qw(cluck croak carp confess);
use Exporter;

@Rancid::Login::ISA = qw(Exporter);
@Rancid::Login::EXPORT = qw(open_router open_program set_trace expect_timeout set_verbose send_data get_prompt login_to_router enable_router);


my $VERSION;

# local state
my $_r_trace_io;
my $_r_hostname;
my $_r_expect_timeout= 30;

BEGIN {
	$_r_trace_io= 0;
	1;
}

sub set_trace # 0 or 1
{
	$_r_trace_io= $_[0];
}

sub expect_timeout
{
	$_r_expect_timeout= shift;
}


sub open_program # cmd line in array, 
{
	$_r_hostname= $_[0];  # error messages will show name of command

    my $session= Expect->spawn(@_) || return undef;
	
	$session->log_stdout(0);	# don't show output to stdout
	$session->match_max(1024 * 256);
	return $session;
}

sub open_router # hostname, port
{
	$_r_hostname= shift;
	my $port = shift || 23;
	my $routersocket = IO::Socket::INET->new(PeerAddr => $_r_hostname,
						 PeerPort => $port,
						 Proto    => 'tcp');
	if (! -S $routersocket)
	{
		cluck "$0: cannot open socket to $_r_hostname:$port: $!\n";
		return undef;
	}

	my $session= Expect->exp_init($routersocket);

	$session->match_max(1024 * 256);
	return $session;
}

sub set_verbose # session, value (0, 1, 2)
{
	my $session= shift;
	my $verbosity= shift;
	$session->debug($verbosity);
	$session->exp_internal($verbosity ? 1 : 0);
}

# Fetch a password from standard input.  Turn off echo while
# prompting the user, and restore the old tty mode afterwards.
# Used for debugging.

sub get_password_stdin # prompt,
{
	my $prompt= $_[0];

    print STDERR $prompt;
    my $old_mode=IO::Stty::stty(\*STDIN,'-g');
    # Turn off echoing.
    IO::Stty::stty(\*STDIN,'-echo');
    # Do whatever.. grab input maybe?
    my $read_password = <STDIN>;
    chomp $read_password;
    # Now restore the old mode.
    IO::Stty::stty(\*STDIN,$old_mode);
    print STDERR "\n";
    return $read_password;
}

sub get_prompt  # session,  prompt-regexp, description-for-error
{
	my $session= $_[0];
	my $prompt= $_[1];
	my $description= $_[2] || '"' . $prompt . '"';

	my $rc= $session->expect($_r_expect_timeout, [-re, $prompt]);
	if (!defined($rc))
	{
		my $err= $session->exp_error;
		croak "$0: $_r_hostname: error $err trying to receive $description\n";
	}
	if ($rc == 1)
	{
		warn "$0: $_r_hostname: matched pattern#$rc: $prompt\n" if $_r_trace_io;
		return 1;
	}

	warn "$0: $_r_hostname: no match trying to receive $description\n";
	return undef;
}

sub send_data # session, string w/CR appended if necessary
{
	my $session= $_[0];
	my $sendme= $_[1];

	warn "$0: $_r_hostname: sending { $sendme } ...\n" if $_r_trace_io;
	$session->clear_accum();
	print $session "$sendme";
}

sub login_to_router  #
# $session,$passwd-prompt,$username-prompt,$invalid-match,$valid-match,
#   $enable, $password, $user, $userpassword
{
# Log in to router, using data from cloginrc as needed
	my $session= shift;
	my $passwordprompt= shift || '([Pp]assword|passwd) *:';
	my $usernameprompt= shift || '([Uu]sername|[lL]ogin|[uU]ser name) *:';
	my $invalidmatch= shift || '([iI]nvalid|[iI]ncorrect|[sS]orry)';
	my $validmatch= shift || '# *$';

	# hyphens aren't allowed in the seed as per the RFC, but some folks
	# use 'em anyway.                           v
	my $otpmatch= 'otp-[0-9a-zA-Z]+\s+[0-9]+\s+[-0-9a-zA-Z]+';

    my $rc;
	my $userprompt_seen= 0;
	my ($enable, $password, $user, $userpassword);
	$enable= shift;
	$enable= (defined($enable) ? 1 : 0);
	$password= shift;
	if (!defined($password))
	{ $password= find_rancid_config("password", $_r_hostname, $enable) };

	$user= shift;
	if (!defined($user))
	{ $user= find_rancid_config("user", $_r_hostname, 0) };

	$user= $_ if ($_= find_rancid_config("enauser", $_r_hostname, 0));

	$userpassword= shift;
	if (!defined($userpassword))
	{ $userpassword= find_rancid_config("userpassword", $_r_hostname, 0) };

	while(1)
	{
		$rc= $session->expect($_r_expect_timeout,
			[-re, $validmatch],
			[-re, $invalidmatch],
			[-re, $otpmatch],
			[-re, $passwordprompt],
			[-re, $usernameprompt],
		);

		my $err= $session->exp_error;
		die "$0: $_r_hostname: error $err awaiting password response from host\n" if !defined($rc);

		if ($rc==1) {return 1}; # valid match - we're logged in
		if ($rc==2) {return 0}; # invalid match - user/password failure
		if ($rc==3) { # otpmatch
			my $challenge= $session->exp_match;
			$userpassword= `otphelper $_r_hostname $challenge`;
			chomp $userpassword;
			die "$0: $_r_hostname: failure running 'otphelper $_r_hostname $challenge'\nRun 'otphelper' at command line for full diagnostics.\n" unless ($? >> 8) == 0;
			# If the password prompt has already been received, fall thru:
			$rc=4 if $session->exp_after =~ /$passwordprompt/;
		}
		if ($rc==4) { # passwordprompt
			send_data($session,
				($userprompt_seen ? $userpassword : $password) . "\n");
		}
		if ($rc==5) { # usernameprompt
			 send_data($session, $user . "\n"); $userprompt_seen= 1;
		}
	}
}

# $session, initial-command, $passwd-prompt, $username-prompt, $invalid-match, $valid-match
sub enable_router #
{
	my $session= shift;
	my $initial_command= shift;
	send_data($session, $initial_command) if defined($initial_command);
	unshift @_, $session;
	push @_, 1;
	return login_to_router(@_);
}

1;
